﻿/*****************************************************************************************************
 * NREL's Solar Position Algorithm (SPA)
 * This algorithm calculates the solar zenith and azimuth angles in the period from the year -2000 to 6000,
 * with uncertainties of +/- 0.0003 degrees based on the date, time, and location on Earth.
 * (Reference: Reda, I.; Andreas, A., Solar Position Algorithm for Solar Radiation Applications, 
 * Solar Energy. Vol. 76(5), 2004; pp. 577-589). The software has not been tested on a variety of 
 * platforms and is not guaranteed to work on yours. It is provided here as a convenience.
 *
 *  Converted from http://code.ascend4.org/ascend/trunk/models/johnpye/ to C# and JavaScript  
 *      apr 2015, Ewald Kendziorra sr, http://www.kendziorra.nl                                    
 *****************************************************************************************************
 * basic usage:
      // fill in the input data "spaIn.<element>" (see structure "strucIn" for explanation)
        ....
      spaIn.function = 2;  // see "const int <calculation type>" for sunpos.spaIn.function
      int res = spa_calculate();  // calculate solar position, sunrise, sun transit, sunset
      if (res == 0)
      {
        // get the SPA-data from "spaOut.<element>" (see structure "strucOut" for explanation)
        ....
      }
      else
      {
        // ERROR (see function "validate_inputs()" for error codes)
      }  
******************************************************************************************************/

/************************************************************
 * combined text from original spaDat.h and spaDat.c files
 ************************************************************/ 

/*
THIS CODE COMES FROM APPENDIX A.6 OF REPORT NREL/TP-560-34302 
WHICH WAS OBTAINED FROM
http://www.osti.gov/bridge/servlets/purl/15003974-iP3z6k/native/15003974.PDF

THE CODE HAS BEEN MODIFIED TO
(1) convert local functions to 'static' to reduce symbol exports in object code
(2) add functionality to allow Julian Day as input, instead of datatime fields.
*/

/////////////////////////////////////////////
// HEADER FILE for spaDat.C //
// //
// Solar Position Algorithm (SPA) //
// for //
// Solar Radiation Application //
// //
// May 12, 2003 //
// //
// Filename: spaDat.H //
// //
// Afshin Michael Andreas //
// afshin_andreas@nrel.gov (303)384-6383 //
// //
// Measurement & Instrumentation Team //
// Solar Radiation Research Laboratory //
// National Renewable Energy Laboratory //
// 1617 Cole Blvd, Golden, CO 80401 //
////////////////////////////////////////////////////////////////////////
// //
// Usage: //
// //
// 1) In calling program, include this header file, //
// by adding this line to the top of file: //
// #include "spaDat.h" //
// //
// 2) In calling program, declare the SPA structure: //
// spa_data spa; //
// //
// 3) Enter the required input values into SPA structure //
// (input values listed in comments below) //
// //
// 4) Call the SPA calculate function and pass the SPA structure //
// (prototype is declared at the end of this header file): //
// spa_calculate(&spa); //
// //
// Selected output values (listed in comments below) will be //
// computed and returned in the passed SPA structure. Output //
// will based on function code selected from enumeration below. //
// //
// Note: A non-zero return code from spa_calculate() indicates that //
// one of the input values did not pass simple bounds tests. //
// The valid input ranges and return error codes are also //
// listed below. //
// //
////////////////////////////////////////////////////////////////////////
// See the spaDat.H header file for usage //
// //
// This code is based on the NREL //
// technical report "Solar Position //
// Algorithm for Solar Radiation //
// Application" by I. Reda & A. Andreas //
///////////////////////////////////////////////////////////////////////////////////////////////
//
// NOTICE
//
//This solar position algorithm for solar radiation applications (the "data") was produced by
//the National Renewable Energy Laboratory ("NREL"), which is operated by the Midwest Research
//Institute ("MRI") under Contract No. DE-AC36-99-GO10337 with the U.S. Department of Energy
//(the "Government").
//
//Reference herein, directly or indirectly to any specific commercial product, process, or
//service by trade name, trademark, manufacturer, or otherwise, does not constitute or imply
//its endorsement, recommendation, or favoring by the Government, MRI or NREL.
//
//THESE DATA ARE PROVIDED "AS IS" AND NEITHER THE GOVERNMENT, MRI, NREL NOR ANY OF THEIR
//EMPLOYEES, MAKES ANY WARRANTY, EXPRESS OR IMPLIED, INCLUDING THE WARRANTIES OF MERCHANTABILITY
//AND FITNESS FOR A PARTICULAR PURPOSE, OR ASSUMES ANY LEGAL LIABILITY OR RESPONSIBILITY FOR THE
//ACCURACY, COMPLETENESS, OR USEFULNESS OF ANY SUCH INFORMATION DISCLOSED IN THE ALGORITHM, OR
//OF ANY APPARATUS, PRODUCT, OR PROCESS DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE
//PRIVATELY OWNED RIGHTS.
///////////////////////////////////////////////////////////////////////////////////////////////
// Revised 27-FEB-2004 Andreas
// Added bounds check on inputs and return value for spa_calculate().
// Revised 10-MAY-2004 Andreas
// Changed temperature bound check minimum from -273.15 to -273 degrees C.
// Revised 17-JUN-2004 Andreas
// Corrected a problem that caused a bogus sunrise/set/transit on the equinox.
// Revised 18-JUN-2004 Andreas
// Added a "function" input variable that allows the selecting of desired outputs.
// Revised 21-JUN-2004 Andreas
// Added 3 new intermediate output values to SPA structure (srha, ssha, & sta).
// Revised 23-JUN-2004 Andreas
// Enumerations for "function" were renamed and 2 were added.
// Prevented bound checks on inputs that are not used (based on function).
// Revised 01-SEP-2004 Andreas
// Changed a local variable from integer to double.
// Revised 12-JUL-2005 Andreas
// Put a limit on the EOT calculation, so that the result is between -20 and 20.
// Revised 26-OCT-2005 Andreas
// Set the atmos. refraction correction to zero, when sun is below horizon.
// Made atmos_refract input a requirement for all "functions".
// Changed atmos_refract bound check from +/- 10 to +/- 5 degrees.
// Revised 07-NOV-2006 Andreas
// Corrected 3 earth periodic terms in the L_TERMS array.
// Corrected 2 earth periodic terms in the R_TERMS array.
// Revised 10-NOV-2006 Andreas
// Corrected a constant used to calculate topocentric sun declination.
// Put a limit on observer hour angle, so result is between 0 and 360.
// Revised 13-NOV-2006 Andreas
// Corrected calculation of topocentric sun declination.
// Converted all floating point inputs in spa structure to doubles.
// Revised 27-FEB-2007 Andreas
// Minor correction made as to when atmos. refraction correction is set to zero.
// Revised 21-JAN-2008 Andreas
// Minor change to two variable declarations.
// Revised 12-JAN-2009 Andreas
// Changed timezone bound check from +/-12 to +/-18 hours.
// Revised 14-JAN-2009 Andreas
// Corrected a constant used to calculate ecliptic mean obliquity.
///////////////////////////////////////////////////////////////////////////////////////////////

const PI = 3.1415926535897932384626433832795028841971;
const SUN_RADIUS = 0.26667;
const SPA_ZA = 0; //calculate zenith and azimuth
const SPA_ZA_INC = 1; //calculate zenith, azimuth, and incidence
const SPA_ZA_RTS = 2; //calculate zenith, azimuth, and sun rise/transit/set values
const SPA_JD = 3; // just calculate the julian day and return
const SPA_ZA_JD = 4; // as SPA_ZA but don't calculate julian day afresh
const SPA_ZA_INC_JD = 5; // as SPA_ZA_INC but don't calculate julian day afresh
const SPA_ZA_RTS_JD = 6; // as SPA_ZA_RTS but don't calculate julian day afresh
const SPA_ALL = 7; // calculate all SPA output values from datetime inputs

/*
    public struct strucIn
    {  //----------------------INPUT VALUES------------------------
      public int year;            // 4-digit year,    valid range: -2000 to 6000, error code: 1
      public int month;           // 2-digit month,         valid range: 1 to 12, error code: 2
      public int day;             // 2-digit day,           valid range: 1 to 31, error code: 3
      public int hour;            // Observer local hour,   valid range: 0 to 24, error code: 4
      public int minute;          // Observer local minute, valid range: 0 to 59, error code: 5
      public int second;          // Observer local second, valid range: 0 to 59, error code: 6
      public double delta_t;      // Difference between earth rotation time and terrestrial time, error code: 7
      public double timezone;     // Observer time zone (negative west of Greenwich) valid range: -18   to   18 hours,   error code: 8
      public double longitude;    // Observer longitude (negative west of Greenwich) valid range: -180  to  180 degrees, error code: 9
      public double latitude;     // Observer latitude (negative south of equator) valid range: -90   to   90 degrees, error code: 10
      public double elevation;    // Observer elevation [meters] valid range: -6500000 or higher meters,    error code: 11
      public double pressure;     // Annual average local pressure [millibars] valid range:    0 to 5000 millibars,       error code: 12
      public double temperature;  // Annual average local temperature [degrees Celsius] valid range: -273 to 6000 degrees Celsius, error code; 13
      public double slope;        // Surface slope (measured from the horizontal plane) valid range: -360 to 360 degrees, error code: 14
      public double azm_rotation; // Surface azimuth rotation (measured from south to projection of surface normal on horizontal plane, negative west) valid range: -360 to 360 degrees, error code: 15
      public double atmos_refract;// Atmospheric refraction at sunrise and sunset (0.5667 deg is typical) valid range: -5   to   5 degrees, error code: 16
      public int function;        // Switch to choose functions for desired output (from enumeration)
    }; public strucIn spaIn = new strucIn();

    public struct strucOut
    {  //---------------------Final OUTPUT VALUES------------------------
      public double zenith;      //topocentric zenith angle [degrees]
      public double azimuth180;  //topocentric azimuth angle (westward from south) [-180 to 180 degrees]
      public double azimuth;     //topocentric azimuth angle (eastward from north) [   0 to 360 degrees]
      public double incidence;   //surface incidence angle [degrees]
      public double suntransit;  //local sun transit time (or solar noon) [fractional hour]
      public double sunrise;     //local sunrise time (+/- 30 seconds) [fractional hour]
      public double sunset;      //local sunset time (+/- 30 seconds) [fractional hour]
    }; public strucOut spaOut = new strucOut();
*/

var spaIn = 
{
  "year": 0,"month": 0,"day": 0,"hour": 0,"minute": 0,"second": 0,"delta_t": 0,"timezone": 0,"longitude": 0,"latitude": 0,
  "elevation": 0,"pressure": 0,"temperature": 0,"slope": 0,"azm_rotation": 0,"atmos_refract": 0,"function": 0
  };
var spaOut = 
{
  "zenith": 0,"azimuth180": 0,"azimuth": 0,"incidence": 0,"suntransit": 0,"sunrise": 0,"sunset": 0
};
var spaDat = 
{
  "jd": 0,"jc": 0,
  "jde": 0,"jce": 0,"jme": 0,"l": 0,"b": 0,"r": 0,"theta": 0,"beta": 0,"x0": 0,"x1": 0,"x2": 0,"x3": 0,"x4": 0,"del_psi": 0,
  "del_epsilon": 0,"epsilon0": 0,"epsilon": 0,"del_tau": 0,"lamda": 0,"nu0": 0,"nu": 0,"alpha": 0,"delta": 0,"h": 0,"xi": 0,
  "del_alpha": 0,"delta_prime": 0,"alpha_prime": 0,"h_prime": 0,"e0": 0,"del_e": 0,"e": 0,"eot": 0,"srha": 0,"ssha": 0,"sta": 0
};

const L_COUNT = 6, B_COUNT = 2, R_COUNT = 5, Y_COUNT = 63;
const TERM_A = 0, TERM_B = 1, TERM_C = 2;
const TERM_X0 = 0, TERM_X1 = 1, TERM_X2 = 2, TERM_X3 = 3, TERM_X4 = 4, TERM_X_COUNT = 5;
const TERM_PSI_A = 0, TERM_PSI_B = 1, TERM_EPS_C = 2, TERM_EPS_D = 3;
const JD_MINUS = 0, JD_ZERO = 1, JD_PLUS = 2, JD_COUNT = 3;
const SUN_TRANSIT = 0, SUN_RISE = 1, SUN_SET = 2, SUN_COUNT = 3;
const TERM_Y_COUNT = TERM_X_COUNT;
const l_subcount = [64, 34, 20, 7, 3, 1];
const b_subcount = [5, 2 ];
const r_subcount = [40, 10, 6, 2, 1];

///////////////////////////////////////////////////
/// Earth Periodic Terms
///////////////////////////////////////////////////
var L_TERMS_L_COUNT_0 =
[
  [175347046.0,0,0], [3341656.0,4.6692568,6283.07585], [34894.0,4.6261,12566.1517], [3497.0,2.7441,5753.3849], [3418.0,2.8289,3.5231],
  [3136.0,3.6277,77713.7715], [2676.0,4.4181,7860.4194], [2343.0,6.1352,3930.2097], [1324.0,0.7425,11506.7698], [1273.0,2.0371,529.691],
  [1199.0,1.1096,1577.3435], [990,5.233,5884.927], [902,2.045,26.298], [857,3.508,398.149], [780,1.179,5223.694], [753,2.533,5507.553],
  [505,4.583,18849.228], [492,4.205,775.523], [357,2.92,0.067], [317,5.849,11790.629], [284,1.899,796.298], [271,0.315,10977.079],
  [243,0.345,5486.778], [206,4.806,2544.314], [205,1.869,5573.143], [202,2.458,6069.777], [156,0.833,213.299], [132,3.411,2942.463],
  [126,1.083,20.775], [115,0.645,0.98], [103,0.636,4694.003], [102,0.976,15720.839], [102,4.267,7.114], [99,6.21,2146.17], [98,0.68,155.42],
  [86,5.98,161000.69], [85,1.3,6275.96], [85,3.67,71430.7], [80,1.81,17260.15], [79,3.04,12036.46], [75,1.76,5088.63], [74,3.5,3154.69],
  [74,4.68,801.82], [70,0.83,9437.76], [62,3.98,8827.39], [61,1.82,7084.9], [57,2.78,6286.6], [56,4.39,14143.5], [56,3.47,6279.55], [52,0.19,12139.55],
  [52,1.33,1748.02], [51,0.28,5856.48], [49,0.49,1194.45], [41,5.37,8429.24], [41,2.4,19651.05], [39,6.17,10447.39], [37,6.04,10213.29],
  [37,2.57,1059.38], [36,1.71,2352.87], [36,1.78,6812.77], [33,0.59,17789.85], [30,0.44,83996.85], [30,2.74,1349.87], [25,3.16,4690.48]
];
var L_TERMS_L_COUNT_1 =
[
  [628331966747.0,0,0], [206059.0,2.678235,6283.07585], [4303.0,2.6351,12566.1517], [425.0,1.59,3.523], [119.0,5.796,26.298], [109.0,2.966,1577.344],
  [93,2.59,18849.23], [72,1.14,529.69], [68,1.87,398.15], [67,4.41,5507.55], [59,2.89,5223.69], [56,2.17,155.42], [45,0.4,796.3], [36,0.47,775.52], 
  [29,2.65,7.11], [21,5.34,0.98], [19,1.85,5486.78], [19,4.97,213.3], [17,2.99,6275.96], [16,0.03,2544.31], [16,1.43,2146.17], [15,1.21,10977.08],
  [12,2.83,1748.02], [12,3.26,5088.63], [12,5.27,1194.45], [12,2.08,4694], [11,0.77,553.57], [10,1.3,6286.6], [10,4.24,1349.87], [9,2.7,242.73],
  [9,5.64,951.72], [8,5.3,2352.87], [6,2.65,9437.76], [6,4.67,4690.48]
];
var L_TERMS_L_COUNT_2 =
[
  [52919.0,0,0], [8720.0,1.0721,6283.0758], [309.0,0.867,12566.152], [27,0.05,3.52], [16,5.19,26.3], [16,3.68,155.42], [10,0.76,18849.23],
  [9,2.06,77713.77], [7,0.83,775.52], [5,4.66,1577.34], [4,1.03,7.11], [4,3.44,5573.14], [3,5.14,796.3], [3,6.05,5507.55], [3,1.19,242.73],
  [3,6.12,529.69], [3,0.31,398.15], [3,2.28,553.57], [2,4.38,5223.69], [2,3.75,0.98],
];
var L_TERMS_L_COUNT_3 = [ [ 289.0, 5.844, 6283.076 ], [ 35, 0, 0 ], [ 17, 5.49, 12566.15 ], [ 3, 5.2, 155.42 ], [ 1, 4.72, 3.52 ], [ 1, 5.3, 18849.23 ], [ 1, 5.97, 242.73 ] ];
var L_TERMS_L_COUNT_4 = [ [ 114.0, 3.142, 0 ], [ 8, 4.13, 6283.08 ], [ 1, 3.84, 12566.15 ] ];
var L_TERMS_L_COUNT_5 = [ [ 1, 3.14, 0 ] ];
var B_TERMS_B_COUNT_0 = [ [ 280.0, 3.199, 84334.662 ], [ 102.0, 5.422, 5507.553 ], [ 80, 3.88, 5223.69 ], [ 44, 3.7, 2352.87 ], [ 32, 4, 1577.34 ] ];
var B_TERMS_B_COUNT_1 = [ [ 9, 3.9, 5507.55 ], [ 6, 1.73, 5223.69 ] ];
var R_TERMS_R_COUNT_0 =
[
  [100013989.0,0,0], [1670700.0,3.0984635,6283.07585], [13956.0,3.05525,12566.1517], [3084.0,5.1985,77713.7715], [1628.0,1.1739,5753.3849],
  [1576.0,2.8469,7860.4194], [925.0,5.453,11506.77], [542.0,4.564,3930.21], [472.0,3.661,5884.927], [346.0,0.964,5507.553], [329.0,5.9,5223.694],
  [307.0,0.299,5573.143], [243.0,4.273,11790.629], [212.0,5.847,1577.344], [186.0,5.022,10977.079], [175.0,3.012,18849.228], [110.0,5.055,5486.778],
  [98,0.89,6069.78], [86,5.69,15720.84], [86,1.27,161000.69], [65,0.27,17260.15], [63,0.92,529.69], [57,2.01,83996.85], [56,5.24,71430.7],
  [49,3.25,2544.31], [47,2.58,775.52], [45,5.54,9437.76], [43,6.01,6275.96], [39,5.36,4694], [38,2.39,8827.39], [37,0.83,19651.05], [37,4.9,12139.55],
  [36,1.67,12036.46], [35,1.84,2942.46], [33,0.24,7084.9], [32,0.18,5088.63], [32,1.78,398.15], [28,1.21,6286.6], [28,1.9,6279.55], [26,4.59,10447.39]
];
var R_TERMS_R_COUNT_1 =
[
  [103019.0,1.10749,6283.07585], [1721.0,1.0644,12566.1517], [702.0,3.142,0], [32,1.02,18849.23], [31,2.84,5507.55], [25,1.32,5223.69],
  [18,1.42,1577.34], [10,5.91,10977.08], [9,1.42,6275.96], [9,0.27,5486.78]
];
var R_TERMS_R_COUNT_2 = [ [ 4359.0, 5.7846, 6283.0758 ], [ 124.0, 5.579, 12566.152 ], [ 12, 3.14, 0 ], [ 9, 3.63, 77713.77 ], [ 6, 1.87, 5573.14 ], [ 3, 5.47, 18849.23 ] ];
var R_TERMS_R_COUNT_3 = [ [ 145.0, 4.273, 6283.076 ], [ 7, 3.92, 12566.15 ] ];
var R_TERMS_R_COUNT_4 = [ [ 4, 2.56, 6283.08 ] ];

////////////////////////////////////////////////////////////////
/// Periodic Terms for the nutation in longitude and obliquity
////////////////////////////////////////////////////////////////
var Y_TERMS =
[
  [0,0,0,0,1], [-2,0,0,2,2], [0,0,0,2,2], [0,0,0,0,2], [0,1,0,0,0], [0,0,1,0,0], [-2,1,0,2,2], [0,0,0,2,1], [0,0,1,2,2], [-2,-1,0,2,2], [-2,0,1,0,0],
  [-2,0,0,2,1], [0,0,-1,2,2], [2,0,0,0,0], [0,0,1,0,1], [2,0,-1,2,2], [0,0,-1,0,1], [0,0,1,2,1], [-2,0,2,0,0], [0,0,-2,2,1], [2,0,0,2,2], [0,0,2,2,2],
  [0,0,2,0,0], [-2,0,1,2,2], [0,0,0,2,0], [-2,0,0,2,0], [0,0,-1,2,1], [0,2,0,0,0], [2,0,-1,0,1], [-2,2,0,2,2], [0,1,0,0,1], [-2,0,1,0,1], [0,-1,0,0,1],
  [0,0,2,-2,0], [2,0,-1,2,1], [2,0,1,2,2], [0,1,0,2,2], [-2,1,1,0,0], [0,-1,0,2,2], [2,0,0,2,1], [2,0,1,0,0], [-2,0,2,2,2], [-2,0,1,2,1], [2,0,-2,0,1],
  [2,0,0,0,1], [0,-1,1,0,0], [-2,-1,0,2,1], [-2,0,0,0,1], [0,0,2,2,1], [-2,0,2,0,1], [-2,1,0,2,1], [0,0,1,-2,0], [-1,0,1,0,0], [-2,1,0,0,0], [1,0,0,0,0],
  [0,0,1,2,0], [0,0,-2,2,2], [-1,-1,1,0,0], [0,1,1,0,0], [0,-1,1,2,2], [2,-1,-1,2,2], [0,0,3,2,2], [2,-1,0,2,2],
];
var PE_TERMS =
[
  [-171996,-174.2,92025,8.9], [-13187,-1.6,5736,-3.1], [-2274,-0.2,977,-0.5], [2062,0.2,-895,0.5], [1426,-3.4,54,-0.1], [712,0.1,-7,0], [-517,1.2,224,-0.6],
  [-386,-0.4,200,0], [-301,0,129,-0.1], [217,-0.5,-95,0.3], [-158,0,0,0], [129,0.1,-70,0], [123,0,-53,0], [63,0,0,0], [63,0.1,-33,0], [-59,0,26,0],
  [-58,-0.1,32,0], [-51,0,27,0], [48,0,0,0], [46,0,-24,0], [-38,0,16,0], [-31,0,13,0], [29,0,0,0], [29,0,-12,0], [26,0,0,0], [-22,0,0,0], [21,0,-10,0],
  [17,-0.1,0,0], [16,0,-8,0], [-16,0.1,7,0], [-15,0,9,0], [-13,0,7,0], [-12,0,6,0], [11,0,0,0], [-10,0,5,0], [-8,0,3,0], [7,0,-3,0], [-7,0,0,0], [-7,0,3,0],
  [-7,0,3,0], [6,0,0,0], [6,0,-3,0], [6,0,-3,0], [-6,0,3,0], [-6,0,3,0], [5,0,0,0], [-5,0,3,0], [-5,0,3,0], [-5,0,3,0], [4,0,0,0], [4,0,0,0], [4,0,0,0],
  [-4,0,0,0], [-4,0,0,0], [-4,0,0,0], [3,0,0,0], [-3,0,0,0], [-3,0,0,0], [-3,0,0,0], [-3,0,0,0], [-3,0,0,0], [-3,0,0,0], [-3,0,0,0],
];
///////////////////////////////////////////////

function rad2deg(radians) { return (180.0 / PI) * radians; }
function deg2rad(degrees) { return (PI / 180.0) * degrees; }

function limit_degrees(degrees)
{
  var limited;
  degrees /= 360.0;
  limited = 360.0 * (degrees - Math.floor(degrees));
  if (limited < 0) limited += 360.0;
  return limited;
}

function limit_degrees180pm(degrees)
{
  var limited;
  degrees /= 360.0;
  limited = 360.0 * (degrees - Math.floor(degrees));
  if (limited < -180.0) limited += 360.0;
  else if (limited > 180.0) limited -= 360.0;
  return limited;
}

function limit_degrees180(degrees)
{
  var limited;
  degrees /= 180.0;
  limited = 180.0 * (degrees - Math.floor(degrees));
  if (limited < 0) limited += 180.0;
  return limited;
}

function limit_zero2one(value)
{
  var limited;
  limited = value - Math.floor(value);
  if (limited < 0) limited += 1.0;
  return limited;
}

function limit_minutes(minutes)
{
  var limited = minutes;
  if (limited < -20.0) limited += 1440.0;
  else if (limited > 20.0) limited -= 1440.0;
  return limited;
}

function dayfrac_to_local_hr(dayfrac)
{
  return 24.0 * limit_zero2one(dayfrac + spaIn.timezone / 24.0);
}

function third_order_polynomial(a, b, c, d)
{
  return ((a * spaDat.jce + b) * spaDat.jce + c) * spaDat.jce + d;
}

function validate_inputs()
{
  switch (spaIn.function)
  {
    case SPA_ZA_JD:
    case SPA_ZA_INC_JD:
    case SPA_ZA_RTS_JD:
      if ((spaDat.jd < 990557.5) || (spaDat.jd > 3912879.500000)) return 16;
      break;
    default:
      if ((spaIn.year < -2000) || (spaIn.year > 6000)) return 1;
      if ((spaIn.month < 1) || (spaIn.month > 12)) return 2;
      if ((spaIn.day < 1) || (spaIn.day > 31)) return 3;
      if ((spaIn.hour < 0) || (spaIn.hour > 24)) return 4;
      if ((spaIn.minute < 0) || (spaIn.minute > 59)) return 5;
      if ((spaIn.second < 0) || (spaIn.second > 59)) return 6;
      break;
  }

  if ((spaIn.pressure < 0) || (spaIn.pressure > 5000)) return 12;
  if ((spaIn.temperature <= -273) || (spaIn.temperature > 6000)) return 13;
  if ((spaIn.hour == 24) && (spaIn.minute > 0)) return 5;
  if ((spaIn.hour == 24) && (spaIn.second > 0)) return 6;
  if (Math.abs(spaIn.delta_t) > 8000) return 7;
  if (Math.abs(spaIn.timezone) > 18) return 8;
  if (Math.abs(spaIn.longitude) > 180) return 9;
  if (Math.abs(spaIn.latitude) > 90) return 10;
  if (Math.abs(spaIn.atmos_refract) > 5) return 16;
  if (spaIn.elevation < -6500000) return 11;
  if ((spaIn.function == SPA_ZA_INC) || (spaIn.function == SPA_ALL))
  {
    if (Math.abs(spaIn.slope) > 360) return 14;
    if (Math.abs(spaIn.azm_rotation) > 360) return 15;
  }
  return 0;
}

function julian_day()
{
  var day_decimal, julian_day, a;
  day_decimal = spaIn.day + (spaIn.hour - spaIn.timezone + (spaIn.minute + spaIn.second / 60.0) / 60.0) / 24.0;
  if (spaIn.month < 3)
  {
    spaIn.month += 12;
    spaIn.year--;
  }
  julian_day = Math.floor(365.25 * (spaIn.year + 4716.0)) + Math.floor(30.6001 * (spaIn.month + 1)) + day_decimal - 1524.5;
  if (julian_day > 2299160.0)
  {
    a = Math.floor(spaIn.year / 100);
    julian_day += (2 - a + Math.floor(a / 4));
  }
  return julian_day;
}

function earth_periodic_term_summation(terms, count)
{
  var i;
  var sum = 0;
  for (i = 0; i < count; i++)
    sum += terms[i][TERM_A] * Math.cos(terms[i][TERM_B] + terms[i][TERM_C] * spaDat.jme);
  return sum;
}

function earth_values(term_sum, count)
{
  var i;
  var sum = 0;
  for (i = 0; i < count; i++)
    sum += term_sum[i] * Math.pow(spaDat.jme, i);
  sum /= 1.0e8;
  return sum;
}

function earth_heliocentric_longitude()
{
  var sum = new Array(L_COUNT);
  sum[0] = earth_periodic_term_summation(L_TERMS_L_COUNT_0, l_subcount[0]);
  sum[1] = earth_periodic_term_summation(L_TERMS_L_COUNT_1, l_subcount[1]);
  sum[2] = earth_periodic_term_summation(L_TERMS_L_COUNT_2, l_subcount[2]);
  sum[3] = earth_periodic_term_summation(L_TERMS_L_COUNT_3, l_subcount[3]);
  sum[4] = earth_periodic_term_summation(L_TERMS_L_COUNT_4, l_subcount[4]);
  sum[5] = earth_periodic_term_summation(L_TERMS_L_COUNT_5, l_subcount[5]);
  spaDat.l = limit_degrees(rad2deg(earth_values(sum, L_COUNT)));
}

function earth_heliocentric_latitude()
{
  var sum = new Array(B_COUNT);
  sum[0] = earth_periodic_term_summation(B_TERMS_B_COUNT_0, b_subcount[0]);
  sum[1] = earth_periodic_term_summation(B_TERMS_B_COUNT_0, b_subcount[1]);
  spaDat.b = rad2deg(earth_values(sum, B_COUNT));
}

function earth_radius_vector()
{
  var sum = new Array(R_COUNT);
  sum[0] = earth_periodic_term_summation(R_TERMS_R_COUNT_0, r_subcount[0]);
  sum[1] = earth_periodic_term_summation(R_TERMS_R_COUNT_1, r_subcount[1]);
  sum[2] = earth_periodic_term_summation(R_TERMS_R_COUNT_2, r_subcount[2]);
  sum[3] = earth_periodic_term_summation(R_TERMS_R_COUNT_3, r_subcount[3]);
  sum[4] = earth_periodic_term_summation(R_TERMS_R_COUNT_4, r_subcount[4]);
  spaDat.r = earth_values(sum, R_COUNT);
}

function xy_term_summation(i, x)
{
  var j;
  var sum = 0;
  for (j = 0; j < TERM_Y_COUNT; j++) sum += x[j] * Y_TERMS[i][j];
  return sum;
}

function nutation_longitude_and_obliquity(x)
{
  var i;
  var xy_term_sum, sum_psi = 0, sum_epsilon = 0;
  for (i = 0; i < Y_COUNT; i++)
  {
    xy_term_sum = deg2rad(xy_term_summation(i, x));
    sum_psi += (PE_TERMS[i][TERM_PSI_A] + spaDat.jce * PE_TERMS[i][TERM_PSI_B]) * Math.sin(xy_term_sum);
    sum_epsilon += (PE_TERMS[i][TERM_EPS_C] + spaDat.jce * PE_TERMS[i][TERM_EPS_D]) * Math.cos(xy_term_sum);
  }
  spaDat.del_psi = sum_psi / 36000000.0;
  spaDat.del_epsilon = sum_epsilon / 36000000.0;
}

function sun_right_ascension_parallax_and_topocentric_dec()
{
  var delta_alpha_rad;
  var lat_rad = deg2rad(spaIn.latitude);
  var xi_rad = deg2rad(spaDat.xi);
  var h_rad = deg2rad(spaDat.h);
  var delta_rad = deg2rad(spaDat.delta);
  var u = Math.atan(0.99664719 * Math.tan(lat_rad));
  var y = 0.99664719 * Math.sin(u) + spaIn.elevation * Math.sin(lat_rad) / 6378140.0;
  var x = Math.cos(u) + spaIn.elevation * Math.cos(lat_rad) / 6378140.0;
  delta_alpha_rad = Math.atan2(-x * Math.sin(xi_rad) * Math.sin(h_rad), Math.cos(delta_rad) - x * Math.sin(xi_rad) * Math.cos(h_rad));
  spaDat.delta_prime = rad2deg(Math.atan2((Math.sin(delta_rad) - y * Math.sin(xi_rad)) * Math.cos(delta_alpha_rad), Math.cos(delta_rad) - x * Math.sin(xi_rad) * Math.cos(h_rad)));
  spaDat.del_alpha = rad2deg(delta_alpha_rad);
}

function rts_alpha_delta_prime(ad, n)
{
  var a = ad[JD_ZERO] - ad[JD_MINUS];
  var b = ad[JD_PLUS] - ad[JD_ZERO];
  if (Math.abs(a) >= 2.0) a = limit_zero2one(a);
  if (Math.abs(b) >= 2.0) b = limit_zero2one(b);
  return ad[JD_ZERO] + n * (a + b + (b - a) * n) / 2.0;
}

function sun_rise_and_set(m_rts, h_rts, delta_prime, h_prime, h0_prime, sun)
{
  return m_rts[sun] + (h_rts[sun] - h0_prime) / (360.0 * Math.cos(deg2rad(delta_prime[sun])) * Math.cos(deg2rad(spaIn.latitude)) * Math.sin(deg2rad(h_prime[sun])));
}

////////////////////////////////////////////////////////////////////////////////////////////////
// Calculate required SPA parameters to get the right ascension (alpha) and declination (delta)
// Note: JD must be already calculated and in structure
////////////////////////////////////////////////////////////////////////////////////////////////
function calculate_geocentric_sun_right_ascension_and_declination()
{
  var u, lamda_rad, epsilon_rad, beta_rad;
  var x = new Array(TERM_X_COUNT);
  spaDat.jc = (spaDat.jd - 2451545.0) / 36525.0;  // julian_century
  spaDat.jde = spaDat.jd + spaIn.delta_t / 86400.0;  // julian_ephemeris_day
  spaDat.jce = (spaDat.jde - 2451545.0) / 36525.0;  // julian_ephemeris_century
  spaDat.jme = (spaDat.jce / 10.0);  // ulian_ephemeris_millennium
  earth_heliocentric_longitude();
  earth_heliocentric_latitude();
  earth_radius_vector();
  spaDat.theta = spaDat.l + 180.0;  // geocentric_longitude
  spaDat.beta = -spaDat.b;  // geocentric_latitude
  x[TERM_X0] = spaDat.x0 = third_order_polynomial(1.0 / 189474.0, -0.0019142, 445267.11148, 297.85036);  // mean_elongation_moon_sun
  x[TERM_X1] = spaDat.x1 = third_order_polynomial(-1.0 / 300000.0, -0.0001603, 35999.05034, 357.52772);  // mean_anomaly_sun
  x[TERM_X2] = spaDat.x2 = third_order_polynomial(1.0 / 56250.0, 0.0086972, 477198.867398, 134.96298);  // mean_anomaly_moon
  x[TERM_X3] = spaDat.x3 = third_order_polynomial(1.0 / 327270.0, -0.0036825, 483202.017538, 93.27191);  // argument_latitude_moon
  x[TERM_X4] = spaDat.x4 = third_order_polynomial(1.0 / 450000.0, 0.0020708, -1934.136261, 125.04452);  // ascending_longitude_moon
  nutation_longitude_and_obliquity(x);
  u = spaDat.jme / 10.0;
  spaDat.epsilon0 = 84381.448 + u * (-4680.93 + u * (-1.55 + u * (1999.25 + u * (-51.38 + u * (-249.67 + u * (-39.05 + u * (7.12 + u * (27.87 + u * (5.79 + u * 2.45)))))))));  // ecliptic_mean_obliquity
  spaDat.epsilon = spaDat.del_epsilon + spaDat.epsilon0 / 3600.0;  // ecliptic_true_obliquity
  spaDat.del_tau = -20.4898 / (3600.0 * spaDat.r);  // aberration_correction
  spaDat.lamda = spaDat.theta + spaDat.del_psi + spaDat.del_tau;  // apparent_sun_longitude
  spaDat.nu0 = limit_degrees(280.46061837 + 360.98564736629 * (spaDat.jd - 2451545.0) + spaDat.jc * spaDat.jc * (0.000387933 - spaDat.jc / 38710000.0));  // greenwich_mean_sidereal_time
  spaDat.nu = spaDat.nu0 + spaDat.del_psi * Math.cos(deg2rad(spaDat.epsilon));  // greenwich_sidereal_time
  lamda_rad = deg2rad(spaDat.lamda);
  epsilon_rad = deg2rad(spaDat.epsilon);
  spaDat.alpha = limit_degrees(rad2deg(Math.atan2(Math.sin(lamda_rad) * Math.cos(epsilon_rad) - Math.tan(deg2rad(spaDat.beta)) * Math.sin(epsilon_rad), Math.cos(lamda_rad))));  // geocentric_sun_right_ascension
  beta_rad = deg2rad(spaDat.beta);
  epsilon_rad = deg2rad(spaDat.epsilon);
  spaDat.delta = rad2deg(Math.asin(Math.sin(beta_rad) * Math.cos(epsilon_rad) + Math.cos(beta_rad) * Math.sin(epsilon_rad) * Math.sin(deg2rad(spaDat.lamda))));  // geocentric_sun_declination
}

////////////////////////////////////////////////////////////////////////
// Calculate Equation of Time (EOT) and Sun Rise, Transit, & Set (RTS)
////////////////////////////////////////////////////////////////////////

function calculate_eot_and_sun_rise_transit_set()
{
  var nu, m, h0, n;
  var latitude_rad, delta_zero_rad, delta_prime_rad, h0_dfrac, argument;
  var alpha = new Array(JD_COUNT), delta = new Array(JD_COUNT);
  var m_rts = new Array(SUN_COUNT), nu_rts = new Array(SUN_COUNT), h_rts = new Array(SUN_COUNT);
  var alpha_prime = new Array(SUN_COUNT), delta_prime = new Array(SUN_COUNT), h_prime = new Array(SUN_COUNT);
  var h0_prime = -1 * (SUN_RADIUS + spaIn.atmos_refract);
  var i;
  m = limit_degrees(280.4664567 + spaDat.jme * (360007.6982779 + spaDat.jme * (0.03032028 + spaDat.jme * (1 / 49931.0 + spaDat.jme * (-1 / 15300.0 + spaDat.jme * (-1 / 2000000.0))))));  // sun_mean_longitude
  spaDat.eot = limit_minutes(4.0 * (m - 0.0057183 - spaDat.alpha + spaDat.del_psi * Math.cos(deg2rad(spaDat.epsilon))));  // eot 
  spaIn.hour = spaIn.minute = spaIn.second = 0;
  spaIn.timezone = 0.0;
  spaDat.jd = julian_day();
  calculate_geocentric_sun_right_ascension_and_declination();
  nu = spaDat.nu;
  spaIn.delta_t = 0;
  spaDat.jd--;
  for (i = 0; i < JD_COUNT; i++)
  {
    calculate_geocentric_sun_right_ascension_and_declination();
    alpha[i] = spaDat.alpha;
    delta[i] = spaDat.delta;
    spaDat.jd++;
  }
  m_rts[SUN_TRANSIT] = (alpha[JD_ZERO] - spaIn.longitude - nu) / 360.0;  // approx_sun_transit_time
  h0 = -99999;      // sun_hour_angle_at_rise_set
  latitude_rad = deg2rad(spaIn.latitude);
  delta_zero_rad = deg2rad(delta[JD_ZERO]);
  argument = (Math.sin(deg2rad(h0_prime)) - Math.sin(latitude_rad) * Math.sin(delta_zero_rad)) / (Math.cos(latitude_rad) * Math.cos(delta_zero_rad));
  if (Math.abs(argument) <= 1) h0 = limit_degrees180(rad2deg(Math.acos(argument)));
  if (h0 >= 0)
  {
    h0_dfrac = h0 / 360.0;      // approx_sun_rise_and_set
    m_rts[SUN_RISE] = limit_zero2one(m_rts[SUN_TRANSIT] - h0_dfrac);
    m_rts[SUN_SET] = limit_zero2one(m_rts[SUN_TRANSIT] + h0_dfrac);
    m_rts[SUN_TRANSIT] = limit_zero2one(m_rts[SUN_TRANSIT]);
    for (i = 0; i < SUN_COUNT; i++)
    {
      nu_rts[i] = nu + 360.985647 * m_rts[i];
      n = m_rts[i] + spaIn.delta_t / 86400.0;
      alpha_prime[i] = rts_alpha_delta_prime(alpha, n);
      delta_prime[i] = rts_alpha_delta_prime(delta, n);
      h_prime[i] = limit_degrees180pm(nu_rts[i] + spaIn.longitude - alpha_prime[i]);
      latitude_rad = deg2rad(spaIn.latitude);            // rts_sun_altitude
      delta_prime_rad = deg2rad(delta_prime[i]);
      h_rts[i] = rad2deg(Math.asin(Math.sin(latitude_rad) * Math.sin(delta_prime_rad) + Math.cos(latitude_rad) * Math.cos(delta_prime_rad) * Math.cos(deg2rad(h_prime[i]))));
    }
    spaDat.srha = h_prime[SUN_RISE];
    spaDat.ssha = h_prime[SUN_SET];
    spaDat.sta = h_rts[SUN_TRANSIT];
    spaOut.suntransit = dayfrac_to_local_hr(m_rts[SUN_TRANSIT] - h_prime[SUN_TRANSIT] / 360.0);
    spaOut.sunrise = dayfrac_to_local_hr(sun_rise_and_set(m_rts, h_rts, delta_prime, h_prime, h0_prime, SUN_RISE));
    spaOut.sunset = dayfrac_to_local_hr(sun_rise_and_set(m_rts, h_rts, delta_prime, h_prime, h0_prime, SUN_SET));
  }
  else spaDat.srha = spaDat.ssha = spaDat.sta = spaOut.suntransit = spaOut.sunrise = spaOut.sunset = -99999;
}

///////////////////////////////////////////////////////////////////////////////////////////
// Calculate all SPA parameters and put into structure
// Note: All inputs values (listed in header file) must already be in structure
///////////////////////////////////////////////////////////////////////////////////////////
function spa_calculate()
{
var result;
var lat_rad, delta_prime_rad, h_prime_rad, zenith_rad, slope_rad; result = validate_inputs();
if (result == 0)
{
  switch (spaIn.function)
  {
    case SPA_ZA:
    case SPA_ZA_INC:
    case SPA_ZA_RTS:
    case SPA_ALL:
    case SPA_JD:
      spaDat.jd = julian_day();
      if (spaIn.function == SPA_JD)
      {
        return result;
      }
      break;
    default:
      break;
  }
  calculate_geocentric_sun_right_ascension_and_declination();
  spaDat.h = limit_degrees(spaDat.nu + spaIn.longitude - spaDat.alpha);  // observer_hour_angle
  spaDat.xi = 8.794 / (3600.0 * spaDat.r);  // sun_equatorial_horizontal_parallax
  sun_right_ascension_parallax_and_topocentric_dec();
  spaDat.alpha_prime = spaDat.alpha + spaDat.del_alpha;  // topocentric_sun_right_ascension 
  spaDat.h_prime = spaDat.h - spaDat.del_alpha;  // topocentric_local_hour_angle
  lat_rad = deg2rad(spaIn.latitude);
  delta_prime_rad = deg2rad(spaDat.delta_prime);
  spaDat.e0 = rad2deg(Math.asin(Math.sin(lat_rad) * Math.sin(delta_prime_rad) + Math.cos(lat_rad) * Math.cos(delta_prime_rad) * Math.cos(deg2rad(spaDat.h_prime))));  // topocentric_elevation_angle
  spaDat.del_e = 0;
  if (spaDat.e0 >= -1 * (SUN_RADIUS + spaIn.atmos_refract)) spaDat.del_e = (spaIn.pressure / 1010.0) * (283.0 / (273.0 + spaIn.temperature)) * 1.02 / (60.0 * Math.tan(deg2rad(spaDat.e0 + 10.3 / (spaDat.e0 + 5.11))));  // atmospheric_refraction_correction
  spaDat.e = spaDat.e0 + spaDat.del_e;  // topocentric_elevation_angle_corrected
  spaOut.zenith = 90.0 - spaDat.e;  // topocentric_zenith_angle
  h_prime_rad = deg2rad(spaDat.h_prime);           // topocentric_azimuth_angle_neg180_180
  lat_rad = deg2rad(spaIn.latitude);
  spaOut.azimuth180 = rad2deg(Math.atan2(Math.sin(h_prime_rad), Math.cos(h_prime_rad) * Math.sin(lat_rad) - Math.tan(deg2rad(spaDat.delta_prime)) * Math.cos(lat_rad)));
  spaOut.azimuth = spaOut.azimuth180 + 180.0;  // topocentric_azimuth_angle_zero_360
  if ((spaIn.function == SPA_ZA_INC) || (spaIn.function == SPA_ZA_INC_JD) || (spaIn.function == SPA_ALL))
  {
    zenith_rad = deg2rad(spaOut.zenith);
    slope_rad = deg2rad(spaIn.slope);
    spaOut.incidence = rad2deg(Math.acos(Math.cos(zenith_rad) * Math.cos(slope_rad) + Math.sin(slope_rad) * Math.sin(zenith_rad) * Math.cos(deg2rad(spaOut.azimuth180 - spaIn.azm_rotation))));
  }
  if ((spaIn.function == SPA_ZA_RTS) || (spaIn.function == SPA_ZA_RTS_JD) || (spaIn.function == SPA_ALL))
  {
    calculate_eot_and_sun_rise_transit_set();
  }
}
return result;
}
